using System;

namespace MyNamespace
{
	/// <summary>
	///		XIMult - Extra Precision Integer Multiply using Fast Hartley Transform Multiply
	///		Developed in TURBO Pascal 5.5 and 6.0, converted to Visual Basic 6
	///		and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class XIMult
	{
		public XIMult()
		{
			// no constructor logic
		}

		public static void FHTnLdn(int xN, int yN, out int n, out int ldn) // compute n and ldn
		{
			int nMax = 2 * Math.Max(xN, yN);
			ldn = (int)(Math.Ceiling(Math.Log(nMax, 2)));
			n = 1 << ldn;

			if (n >= nMax + nMax)		// for safety, needed if nMax is a power of 2
			{							// and Log(n, 2) = integer + epsilon
				ldn--;					// the only one I could find was n = 2^29, log(n, 2) gave 29.000000000000004 (14 zeros)
				n = 1 << ldn;			// so this may not be needed but better safe ...
			}
		}

		public static int SumOfDigits(MultiUI x) // Sum of the digits of x mod (radix - 1)
		{
			int s = 0;
			for (int i = 0; i < x.N; i++)
			{
				s += x.V[i];
				if (s >= MultiID.maxDigit) s -= MultiID.maxDigit;
			}
			return s;
		}

		public static void XIMul(MultiUI x, MultiUI y, MultiUI z) // z = x * y
		// x or y may NOT have the same address in memory as z
		// x and y may have the same address, they are not changed
		// sets muErr true if overflow
		{
			if(Object.ReferenceEquals(x, y))
			{
				XISquare(x, z);
				return;
			}
			double timeAtStart = MultiID.TimerF();
			int xN = x.N;
			int yN = y.N;
			int n, ldn;
			double[] f;
			double[] g;
			if (MultiID.muDMax != 8)
				FHTnLdn(xN, yN, out n, out ldn);
			else
				FHTnLdn(2*xN, 2*yN, out n, out ldn);
			if (n > (1 << 23)) // 2^23 = 8,388,608
			{
				MultiID.WriteLn("Numbers too large for FHT multiply, n = " + n + " > 8,388,608");
				MultiID.muErr = true;
				MultiID.ClearUI(z);
				return;
			}
			int sodX = SumOfDigits(x);
			int sodY = SumOfDigits(y);
			int sodXY = (int)(Math.BigMul(sodX, sodY) % MultiID.maxDigit);
			try
			{
				f = new double[n];		// initialized to zero
				g = new double[n];
			}
			catch (Exception ex)
			{
				string st = ex.Message + ", n = " + n + ", z set to zero";
				Common.ReportError("XIMul", st);
				MultiID.muErr = true;
				z.N = 1;
				z.M = 1;
				z.V = new int[2];
				return;
			}
			if (MultiID.muDMax != 8)
			{
				for (int i = 0; i < xN; i++)	// copy x, int into double
					f[i] = x.V[i];
				for (int i = 0; i < yN; i++)	// copy y, int into double
					g[i] = y.V[i];
			}
			else
			{
				for (int i = 0; i < xN; i++)	// copy x, int into double
				{
					int rem;
					int q = Math.DivRem(x.V[i], MultiID.baseMr, out rem);
					f[i+i] = rem;
					f[i+i+1] = q;
				}
				for (int i = 0; i < yN; i++)	// copy y, int into double
				{
					int rem;
					int q = Math.DivRem(y.V[i], MultiID.baseMr, out rem);
					g[i+i] = rem;
					g[i+i+1] = q;
				}
			}
			double maxError;
			FHTMult.FHTConvolution0(f, g, ldn);	// do Hartley convolution
			if (MultiID.muAbort)
			{
				MultiID.ClearUI(z);
				return;
			}
			FHTMult.ConvolutionRound(g, ldn, out maxError);
			FHTMult.ConvolutionCarry(g, ldn);
			int zN = xN + yN;
			if (zN > z.M)
			{
				zN = z.M;						// overflow condition, should not happen
				MultiID.WriteLn("FHT overflow, should not happen but it did!!!!!!!!!!!!");
			}
			if (MultiID.muDMax != 8)
			{
				if (g[zN-1] == 0.0d) zN--;
				for (int i = 0; i < zN; i++)
					z.V[i] = (int)g[i];
				z.N = zN;
			}
			else
			{
				for (int i = 0, j = 0;  i < zN;  i++, j += 2)
					z.V[i] = (int)g[j+1] * MultiID.baseMr + (int)g[j];
				if (z.V[zN-1] == 0.0d) zN--;
				z.N = zN;
			}
			int sodZ = SumOfDigits(z);
			if (sodZ != sodXY)
				MultiID.WriteLn("FHT Sum-of-Digits failed, error in FHT multiply!!!!!!!!!!!!");
			if (Math.Abs(maxError) > 0.45)	//error, this should not happen either I hope
				MultiID.WriteLn("FHT Max Error = " + maxError + ", probable error in FHT multiply!!!!!!!!!!!!");
//if (MultiID.diagOn)
//{
//	double delta = MultiID.TimerF() - timeAtStart;
//	if (delta > 0.025)
//	{
//		string st = string.Format("FHT Radix = {0}  ldn = {1}  ", FHTMult.radix, ldn);
//		st += string.Format("Max Error = {0:0.00000}  DT = {1:0.00} sec.  ", maxError, delta);
//		st += string.Format("x.N = {0}  y.N = {1}  z.N = {2}", xN, yN, z.N);
//		MultiID.WriteTDLn(st);
//	}
//}
		} // XIMul

		public static void XISquare(MultiUI x, MultiUI z) // z = x * x
		// x may NOT have the same address in memory as z
		// x is not changed
		// sets muErr true if overflow
		{
			double timeAtStart = MultiID.TimerF();
			int xN = x.N;
			int n, ldn;
			double[] g;
			if (MultiID.muDMax != 8)
				FHTnLdn(xN, xN, out n, out ldn);
			else
				FHTnLdn(2*xN, 2*xN, out n, out ldn);
			if (n > (1 << 23)) // 2^23 = 8,388,608
			{
				MultiID.WriteLn("Numbers too large for FHT multiply, n = " + n + " > 8,388,608");
				MultiID.muErr = true;
				MultiID.ClearUI(z);
				return;
			}
			int sodX = SumOfDigits(x);
			int sodXY = (int)(Math.BigMul(sodX, sodX) % MultiID.maxDigit);
			try
			{
				g = new double[n];		// initialized to zero
			}
			catch (Exception ex)
			{
				string st = ex.Message + ", n = " + n + ", z set to zero";
				Common.ReportError("XIMul", st);
				MultiID.muErr = true;
				z.N = 1;
				z.M = 1;
				z.V = new int[2];
				return;
			}
			if (MultiID.muDMax != 8)
			{
				for (int i = 0; i < xN; i++)	// copy x, int into double
					g[i] = x.V[i];
			}
			else
			{
				for (int i = 0; i < xN; i++)	// copy x, int into double
				{
					int rem;
					int q = Math.DivRem(x.V[i], MultiID.baseMr, out rem);
					g[i+i] = rem;
					g[i+i+1] = q;
				}
			}
			double maxError;
			FHTMult.FHTConvolution0(g, ldn);	// do Hartley convolution of g with g
			if (MultiID.muAbort)
			{
				MultiID.ClearUI(z);
				return;
			}
			FHTMult.ConvolutionRound(g, ldn, out maxError);
			FHTMult.ConvolutionCarry(g, ldn);
			int zN = xN + xN;
			if (zN > z.M)
			{
				zN = z.M;					// overflow condition, should not happen
				MultiID.WriteLn("FHT overflow, should not happen but it did!!!!!!!!!!!!");
			}
			if (MultiID.muDMax != 8)
			{
				if (g[zN-1] == 0.0d) zN--;
				for (int i = 0; i < zN; i++)
					z.V[i] = (int)g[i];
				z.N = zN;
			}
			else
			{
				for (int i = 0, j = 0;  i < zN;  i++, j += 2)
					z.V[i] = (int)g[j+1] * MultiID.baseMr + (int)g[j];
				if (z.V[zN-1] == 0.0d) zN--;
				z.N = zN;
			}
			int sodZ = SumOfDigits(z);
			if (sodZ != sodXY)
				MultiID.WriteLn("FHT Sum-of-Digits failed, error in FHT multiply!!!!!!!!!!!!");
			if (Math.Abs(maxError) > 0.45)	//error, this should not happen either I hope
				MultiID.WriteLn("FHT Max Error = " + maxError + ", probable error in FHT multiply!!!!!!!!!!!!");
		} // XISquare
	}
}
